#![feature(generic_const_exprs)]

use std::{
    f64::consts::PI,
    f64::consts::TAU,
};
use unidades::{
    dimensions::{
        Density,
        Frequency,
        Length,
        Pressure,
        Speed,
    },
    prefixes::{
        Deci,
        Mega,
    },
    units::{
        self,
        KgPerCubicMeter,
        Meter,
        MetersPerSecond,
        Pascal,
        Rpm,
    },
    Unit,
};

pub fn speed_to_rpm(tip_speed: Speed, radius: Length) -> Frequency {
    tip_speed / (radius * TAU)
}

#[derive(Copy, Clone, Debug)]
pub struct Flywheel {
    tensile_strength: Pressure,
    density: Density,
    radius: Length,
    thickness: Length,
}

impl Flywheel {
    pub fn steel(radius: Length, thickness: Length) -> Self {
        Self {
            tensile_strength: Mega::<Pascal>::new(700.0),
            density: KgPerCubicMeter::new(7850.0),
            radius,
            thickness,
        }
    }

    pub fn wood(radius: Length, thickness: Length) -> Self {
        Self {
            tensile_strength: Mega::<Pascal>::new(105.0),
            density: KgPerCubicMeter::new(700.0),
            radius,
            thickness,
        }
    }

    pub fn max_velocity(&self) -> Speed {
        let accel = self.tensile_strength / (self.density * PI.powf(2.0) * self.radius.square());
        MetersPerSecond::new(accel.to_units::<units::MetersPerSecondPerSecond>().powf(0.5))
    }
}

fn main() {
    let steel = Flywheel::steel(Meter::new(0.5), Deci::<Meter>::new(1.0));
    let max_rpm = speed_to_rpm(steel.max_velocity(), steel.radius);
    println!["{steel:?}"];
    println!["{}", max_rpm.to_units::<Rpm>()];
}
