use std::ops::{
    Add,
    Div,
    Mul,
    Sub,
};
use crate::Unit;

#[derive(Copy, Clone, Debug)]
pub struct Measurement<
    const TIME: i8,
    const LENGTH: i8,
    const MASS: i8,
    const CURRENT: i8,
    const TEMPERATURE: i8,
    const AMOUNT: i8,
    const INTENSITY: i8,
>(
    pub f64,
);

impl<
    const TIME: i8,
    const LENGTH: i8,
    const MASS: i8,
    const CURRENT: i8,
    const TEMPERATURE: i8,
    const AMOUNT: i8,
    const INTENSITY: i8,
> Measurement<TIME, LENGTH, MASS, CURRENT, TEMPERATURE, AMOUNT, INTENSITY> {
    pub fn new(v: f64) -> Self {
        Self(v)
    }

    pub fn to_units<U>(self) -> f64
    where
        U: Unit<Self> {
        self.0.sub(U::offset()).div(U::scale())
    }

    pub fn square(self) -> Self {
        Self(self.0.powf(2.0))
    }
}

impl<
    const A_TIME: i8,
    const A_LENGTH: i8,
    const A_MASS: i8,
    const A_CURRENT: i8,
    const A_TEMPERATURE: i8,
    const A_AMOUNT: i8,
    const A_INTENSITY: i8,
    const B_TIME: i8,
    const B_LENGTH: i8,
    const B_MASS: i8,
    const B_CURRENT: i8,
    const B_TEMPERATURE: i8,
    const B_AMOUNT: i8,
    const B_INTENSITY: i8,
> Mul<
    Measurement<B_TIME, B_LENGTH, B_MASS, B_CURRENT, B_TEMPERATURE, B_AMOUNT, B_INTENSITY>,
> for Measurement<A_TIME, A_LENGTH, A_MASS, A_CURRENT, A_TEMPERATURE, A_AMOUNT, A_INTENSITY>
where
    Measurement<{
        A_TIME + B_TIME
    }, {
        A_LENGTH + B_LENGTH
    }, {
        A_MASS + B_MASS
    }, {
        A_CURRENT + B_CURRENT
    }, {
        A_TEMPERATURE + B_TEMPERATURE
    }, {
        A_AMOUNT + B_AMOUNT
    }, {
        A_INTENSITY + B_INTENSITY
    }>: Sized {
    type Output = Measurement<{
        A_TIME + B_TIME
    }, {
        A_LENGTH + B_LENGTH
    }, {
        A_MASS + B_MASS
    }, {
        A_CURRENT + B_CURRENT
    }, {
        A_TEMPERATURE + B_TEMPERATURE
    }, {
        A_AMOUNT + B_AMOUNT
    }, {
        A_INTENSITY + B_INTENSITY
    }>;

    fn mul(
        self,
        rhs: Measurement<B_TIME, B_LENGTH, B_MASS, B_CURRENT, B_TEMPERATURE, B_AMOUNT, B_INTENSITY>,
    ) -> Self::Output {
        Measurement::<{
            A_TIME + B_TIME
        }, {
            A_LENGTH + B_LENGTH
        }, {
            A_MASS + B_MASS
        }, {
            A_CURRENT + B_CURRENT
        }, {
            A_TEMPERATURE + B_TEMPERATURE
        }, {
            A_AMOUNT + B_AMOUNT
        }, {
            A_INTENSITY + B_INTENSITY
        }>(self.0 * rhs.0)
    }
}

impl<
    const TIME: i8,
    const LENGTH: i8,
    const MASS: i8,
    const CURRENT: i8,
    const TEMPERATURE: i8,
    const AMOUNT: i8,
    const INTENSITY: i8,
> Mul<f64> for Measurement<TIME, LENGTH, MASS, CURRENT, TEMPERATURE, AMOUNT, INTENSITY>
where
    Self: Sized {
    type Output = Self;

    fn mul(self, rhs: f64) -> Self::Output {
        Self(self.0 * rhs)
    }
}

impl<
    const TIME: i8,
    const LENGTH: i8,
    const MASS: i8,
    const CURRENT: i8,
    const TEMPERATURE: i8,
    const AMOUNT: i8,
    const INTENSITY: i8,
> Mul<Measurement<TIME, LENGTH, MASS, CURRENT, TEMPERATURE, AMOUNT, INTENSITY>> for f64
where
    Measurement<TIME, LENGTH, MASS, CURRENT, TEMPERATURE, AMOUNT, INTENSITY>: Sized {
    type Output = Measurement<TIME, LENGTH, MASS, CURRENT, TEMPERATURE, AMOUNT, INTENSITY>;

    fn mul(self, rhs: Measurement<TIME, LENGTH, MASS, CURRENT, TEMPERATURE, AMOUNT, INTENSITY>) -> Self::Output {
        Measurement::<TIME, LENGTH, MASS, CURRENT, TEMPERATURE, AMOUNT, INTENSITY>(self * rhs.0)
    }
}

impl<
    const A_TIME: i8,
    const A_LENGTH: i8,
    const A_MASS: i8,
    const A_CURRENT: i8,
    const A_TEMPERATURE: i8,
    const A_AMOUNT: i8,
    const A_INTENSITY: i8,
    const B_TIME: i8,
    const B_LENGTH: i8,
    const B_MASS: i8,
    const B_CURRENT: i8,
    const B_TEMPERATURE: i8,
    const B_AMOUNT: i8,
    const B_INTENSITY: i8,
> Div<
    Measurement<B_TIME, B_LENGTH, B_MASS, B_CURRENT, B_TEMPERATURE, B_AMOUNT, B_INTENSITY>,
> for Measurement<A_TIME, A_LENGTH, A_MASS, A_CURRENT, A_TEMPERATURE, A_AMOUNT, A_INTENSITY>
where
    Measurement<{
        A_TIME - B_TIME
    }, {
        A_LENGTH - B_LENGTH
    }, {
        A_MASS - B_MASS
    }, {
        A_CURRENT - B_CURRENT
    }, {
        A_TEMPERATURE - B_TEMPERATURE
    }, {
        A_AMOUNT - B_AMOUNT
    }, {
        A_INTENSITY - B_INTENSITY
    }>: Sized {
    type Output = Measurement<{
        A_TIME - B_TIME
    }, {
        A_LENGTH - B_LENGTH
    }, {
        A_MASS - B_MASS
    }, {
        A_CURRENT - B_CURRENT
    }, {
        A_TEMPERATURE - B_TEMPERATURE
    }, {
        A_AMOUNT - B_AMOUNT
    }, {
        A_INTENSITY - B_INTENSITY
    }>;

    fn div(
        self,
        rhs: Measurement<B_TIME, B_LENGTH, B_MASS, B_CURRENT, B_TEMPERATURE, B_AMOUNT, B_INTENSITY>,
    ) -> Self::Output {
        Measurement::<{
            A_TIME - B_TIME
        }, {
            A_LENGTH - B_LENGTH
        }, {
            A_MASS - B_MASS
        }, {
            A_CURRENT - B_CURRENT
        }, {
            A_TEMPERATURE - B_TEMPERATURE
        }, {
            A_AMOUNT - B_AMOUNT
        }, {
            A_INTENSITY - B_INTENSITY
        }>(self.0 / rhs.0)
    }
}

impl<
    const TIME: i8,
    const LENGTH: i8,
    const MASS: i8,
    const CURRENT: i8,
    const TEMPERATURE: i8,
    const AMOUNT: i8,
    const INTENSITY: i8,
> Div<f64> for Measurement<TIME, LENGTH, MASS, CURRENT, TEMPERATURE, AMOUNT, INTENSITY>
where
    Self: Sized {
    type Output = Self;

    fn div(self, rhs: f64) -> Self::Output {
        Self(self.0 / rhs)
    }
}

impl<
    const TIME: i8,
    const LENGTH: i8,
    const MASS: i8,
    const CURRENT: i8,
    const TEMPERATURE: i8,
    const AMOUNT: i8,
    const INTENSITY: i8,
> Div<Measurement<TIME, LENGTH, MASS, CURRENT, TEMPERATURE, AMOUNT, INTENSITY>> for f64
where
    Measurement<TIME, LENGTH, MASS, CURRENT, TEMPERATURE, AMOUNT, INTENSITY>: Sized {
    type Output = Measurement<TIME, LENGTH, MASS, CURRENT, TEMPERATURE, AMOUNT, INTENSITY>;

    fn div(self, rhs: Measurement<TIME, LENGTH, MASS, CURRENT, TEMPERATURE, AMOUNT, INTENSITY>) -> Self::Output {
        Measurement::<TIME, LENGTH, MASS, CURRENT, TEMPERATURE, AMOUNT, INTENSITY>(self * rhs.0)
    }
}

impl<
    const TIME: i8,
    const LENGTH: i8,
    const MASS: i8,
    const CURRENT: i8,
    const TEMPERATURE: i8,
    const AMOUNT: i8,
    const INTENSITY: i8,
> Add for Measurement<TIME, LENGTH, MASS, CURRENT, TEMPERATURE, AMOUNT, INTENSITY>
where
    Self: Sized {
    type Output = Self;

    fn add(self, rhs: Self) -> Self::Output {
        Self(self.0 + rhs.0)
    }
}

impl<
    const TIME: i8,
    const LENGTH: i8,
    const MASS: i8,
    const CURRENT: i8,
    const TEMPERATURE: i8,
    const AMOUNT: i8,
    const INTENSITY: i8,
> Sub for Measurement<TIME, LENGTH, MASS, CURRENT, TEMPERATURE, AMOUNT, INTENSITY>
where
    Self: Sized {
    type Output = Self;

    fn sub(self, rhs: Self) -> Self::Output {
        Self(self.0 - rhs.0)
    }
}
