use crate::{
    dimensions::*, prelude::Kilo, Unit
};
use core::ops::{
    Mul,
    Add,
};

const YEAR: f64 = 31_557_600.0;
const DAY: f64 = 60.0 * 60.0 * 24.0;

#[macro_export]
macro_rules! unit{
    ($dimension: ident => $(($nym: ident, ($scale: expr, $offset: expr))), *) => {
        $(pub struct $nym; impl Unit < $dimension > for $nym {
            fn new(v: f64) -> $dimension {
                $dimension:: new(v.mul(Self::scale()).add(Self::offset()))
            }
            fn scale() -> f64 {
                $scale
            }
            fn offset() -> f64 {
                $offset
            }
        }) *
    };
}

unit!{
    Time =>(Second, (1.0, 0.0)),
    (Kalpa, (YEAR * 1e9 * 4.32, 0.0)),
    (Eon, (YEAR * 1e9, 0.0)),
    (GalacticYear, (YEAR * 2.3e8, 0.0)),
    (Age, (YEAR * (2148.0 + 1.0 / 3.0), 0.0)),
    (Millenium, (YEAR * 1000.0, 0.0)),
    (Century, (YEAR * 100.0, 0.0)),
    (Jubilee, (YEAR * 50.0, 0.0)),
    (Indiction, (YEAR * 15.0, 0.0)),
    (Decade, (YEAR * 10.0, 0.0)),
    (Lustrum, (YEAR * 5.0, 0.0)),
    (Olympiad, (YEAR * 4.0, 0.0)),
    (LeapYear, (YEAR + DAY, 0.0)),
    (LunarYear, (DAY * 354.37, 0.0)),
    (Semester, (DAY * 7.0 * 18.0, 0.0)),
    (Quarantine, (DAY * 40.0, 0.0)),
    (Month, (DAY * 30.0, 0.0)),
    (Fortnight, (DAY * 7.0, 0.0)),
    (Week, (60.0e2 * 24.0 * 7.0, 0.0)),
    (Milliday, (DAY * 0.001, 0.0)),
    (Day, (60.0e2 * 24.0, 0.0)),
    (Hour, (60.0e2, 0.0)),
    (Minute, (60.0, 0.0)),
    (Jiffy, (3e-24, 0.0)),
    (Shake, (1e-08, 0.0)),
    (PlanckTime, (5.39e-44, 0.0))
}

unit!{
    Frequency =>(Hertz, (1.0, 0.0)),
    (Rpm, (1.0 / 60.0, 0.0))
}

unit!{
    Length =>(Meter, (1.0, 0.0)),
    (Planck, (1.6e-35, 0.0)),
    (Yard, (0.9144, 0.0)),
    (Feet, (0.3048, 0.0)),
    (Inch, (0.0254, 0.0)),
    (Angstrom, (10e-10, 0.0))
}

unit!{
    Area =>(SqMeter, (1.0, 0.0)),
    (SqPlanck, (1.6e-35, 0.0)),
    (SqYard, (0.9144, 0.0)),
    (SqInch, (0.0254, 0.0)),
    (SqAngstrom, (1e-10, 0.0))
}

unit!{
    Speed =>(MetersPerSecond, (1.0, 0.0)),
    (Kmph, (3_600.0, 0.0)),
    (Mph, (2_237.0, 0.0))
}

unit!{
    Acceleration =>(MetersPerSecondPerSecond, (1.0, 0.0)),
    (Gee, (9.80665, 0.0))
}

unit!{
    Mass =>(Gram, (1e-03, 0.0))
}

pub type Kilogram = Kilo<Gram>;

unit!{
    MassPerDist =>(KgPerMeter, (1.0, 0.0))
}

unit!{
    Density =>(KgPerCubicMeter, (1.0, 0.0))
}

unit!{
    MassMoment =>(KgSqMeter, (1.0, 0.0))
}

unit!{
    Force =>(Newton, (1.0, 0.0))
}

unit!{
    ForceMoment =>(NewtonMeter, (1.0, 0.0))
}

unit!{
    Pressure =>(Pascal, (1.0, 0.0))
}

unit!{
    Energy =>(Joule, (1.0, 0.0)),
    (Watthour, (3600.0, 0.0))
}

unit!{
    Power =>(Watt, (1.0, 0.0))
}

unit!{
    Current =>(Amp, (1.0, 0.0))
}

unit!{
    Charge =>(Coulomb, (1.0, 0.0))
}

unit!{
    Potential =>(Volt, (1.0, 0.0))
}

unit!{
    Capacitance =>(Farad, (1.0, 0.0))
}

unit!{
    Resistance =>(Ohm, (1.0, 0.0))
}

unit!{
    Conductance =>(Siemen, (1.0, 0.0))
}

unit!{
    Flux =>(Weber, (1.0, 0.0))
}

unit!{
    FluxDensity =>(Tesla, (1.0, 0.0))
}

unit!{
    Inductance =>(Henry, (1.0, 0.0))
}

unit!{
    Amount =>
}

unit!{
    Catalytic =>(Katal, (1.0, 0.0))
}

unit!{
    Temperature =>(Kelvin, (1.0, 0.0)),
    (Celsius, (1.0, 273.15)),
    (Fahrenheit, (5.0 / 9.0, 459.67))
}

unit!{
    Intensity =>
}
